package org.kantega.jexmec.devclp;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.ClassLoaderProvider;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;

import static java.util.Collections.singleton;

/**
 * PluginLoader for easy of development of plugins. On startup this PluginLoader will set up monitoring of
 * a list of plugin directories provided through a system property given (by default {@link #DEFAULT_PLUGIN_HOMES_PROPERTY}.
 * For each of these directories the plugin loader will add a classloader that loads resources from the resourcesDirectory
 * (default {@link #DEFAULT_RESOURCES_DIRECTORY}) and classes from the classesDirectory (default {@link #DEFAULT_CLASSES_DIRECTORY}).
 *
 * For each directory this plugin loader will monitor the classes directory for changes. When a changed file is detected the
 * stratgy will remove the old classloader and add a new one. This allows for hot redeployment of single plugins.
 */
public class DevelopmentClassLoaderProvider implements ClassLoaderProvider {

    private static final String DEFAULT_PLUGIN_HOMES_PROPERTY = "jexmec.pluginHomes";
    private static final String DEFAULT_CLASSES_DIRECTORY = "target/classes";
    private static final String DEFAULT_RESOURCES_DIRECTORY = "src/main/resources";
    private String pluginHomesProperty = DEFAULT_PLUGIN_HOMES_PROPERTY;
    private String resourcesDirectory = DEFAULT_RESOURCES_DIRECTORY;
    private String classesDirectory = DEFAULT_CLASSES_DIRECTORY;
    private Thread changeDetector;
    private List<DevelopmentPluginClassLoader> currentClassLoaders = Collections.synchronizedList(new ArrayList<DevelopmentPluginClassLoader>());
    private long changeDetectorInterval = 5000;
    private volatile boolean changeDetectorStopFlag = false;
    private Registry registry;
    private boolean shouldScan = true;
    private ClassLoader parentClassLoader;

    public void setShouldScan(boolean shouldScan) {
        this.shouldScan = shouldScan;
    }

    public void start(Registry registry, ClassLoader parentClassLoader) {
        this.registry = registry;
        this.parentClassLoader = parentClassLoader;
        String homesProperty = System.getProperty(pluginHomesProperty);

        if (homesProperty != null) {
            for (String home : homesProperty.split(",;")) {
                File homeDirectory = new File(home);
                if (homeDirectory.exists()) {
                    File classes = new File(homeDirectory, classesDirectory);
                    File resources = new File(homeDirectory, resourcesDirectory);
                    if (classes.exists()) {

                        final DevelopmentPluginClassLoader loader = createClassLoader(classes, resources, parentClassLoader);
                        currentClassLoaders.add(loader);
                        registry.add(singleton(loader));
                    }

                }
            }

        }
        if(shouldScan && currentClassLoaders.size() > 0) {
            changeDetector = new Thread(new ChangeDetector(), "Jexmec change detector ");
            changeDetector.setDaemon(true);
            changeDetector.start();
        }

    }

    Thread getChangeDetector() {
        return changeDetector;
    }

    public void setChangeDetectorInterval(long changeDetectorInterval) {
        this.changeDetectorInterval = changeDetectorInterval;
    }

    private DevelopmentPluginClassLoader createClassLoader(File classesDirectory, File resourcesDirectory, ClassLoader parentClassLoader) {
        return new DevelopmentPluginClassLoader(classesDirectory, createResourcesDirectory(resourcesDirectory), parentClassLoader);
    }

    /**
     * This method is only here for tests to subclass
     * @param resourcesDirectory the original resource directory
     * @return normally the given directory
     */
    File createResourcesDirectory(File resourcesDirectory) {
        return resourcesDirectory;
    }

    public void stop() {
        changeDetectorStopFlag = true;
        changeDetector.interrupt();
    }

    public void checkForChanges() {
        for (DevelopmentPluginClassLoader loader : new ArrayList<DevelopmentPluginClassLoader>(currentClassLoaders)) {
            if (loader.getLastModified() > loader.getLastCheck()) {
                final DevelopmentPluginClassLoader newLoader = createClassLoader(loader.getClassesDirectory(), loader.getResourcesDirectory(), parentClassLoader);
                registry.replace(singleton(loader), singleton(newLoader));
                currentClassLoaders.remove(loader);
                currentClassLoaders.add(newLoader);
            } else {
                loader.setLastCheck(System.currentTimeMillis());
            }
        }
    }

    private class ChangeDetector implements Runnable {

        public void run() {
            while (true) {
                if(changeDetectorStopFlag) {
                    return;
                }
                checkForChanges();
                try {
                    Thread.sleep(changeDetectorInterval);
                } catch (InterruptedException e) {

                }
            }
        }

    }

    public void setPluginHomesProperty(String pluginHomesProperty) {
        this.pluginHomesProperty = pluginHomesProperty;
    }

    private class DevelopmentPluginClassLoader extends URLClassLoader {

        private URLClassLoader resourcesLoader;

        private long lastCheck = 0;

        private File classesDirectory;
        private File resourcesDirectory;

        private DevelopmentPluginClassLoader(File classesDirectory, File resourcesDirectory, ClassLoader parentClassLoader) {
            super(getUrls(classesDirectory), parentClassLoader);
            this.classesDirectory = classesDirectory;
            this.resourcesDirectory = resourcesDirectory;
            final List<URL> resourceUrls = new ArrayList<URL>();
            try {
                resourceUrls.add(resourcesDirectory.toURI().toURL());
            } catch (MalformedURLException e) {
                throw new RuntimeException(e);
            }

            for(int i = 1; i < getURLs().length; i++) {
                resourceUrls.add(getURLs()[i]);
            }

            resourcesLoader = new URLClassLoader(resourceUrls.toArray(new URL[resourceUrls.size()]), parentClassLoader);
            setLastCheck(getLastModified());
        }


        @Override
        public URL getResource(String name) {
            return resourcesLoader.getResource(name);
        }

        @Override
        public Enumeration<URL> getResources(String name) throws IOException {
            return resourcesLoader.getResources(name);
        }

        @Override
        public InputStream getResourceAsStream(String name) {
            return resourcesLoader.getResourceAsStream(name);
        }

        public void setLastCheck(long lastCheck) {
            this.lastCheck = lastCheck;
        }

        public long getLastCheck() {
            return lastCheck;
        }

        public long getLastModified() {
            return getLastModified(classesDirectory);
        }

        private long getLastModified(File file) {
            if (file.isFile()) {
                return file.lastModified();
            } else {
                long max = 0;
                for (File child : file.listFiles()) {
                    max = Math.max(max, getLastModified(child));
                }
                return max;
            }
        }

        public File getClassesDirectory() {
            return classesDirectory;
        }

        public File getResourcesDirectory() {
            return resourcesDirectory;
        }
    }

    private static URL[] getUrls(File location) {
        try {
            List<URL> urls = new ArrayList<URL>();
            urls.add(location.toURI().toURL());
            File libDir = new File(location, "META-INF/lib");
            if (libDir.exists()) {
                File[] libs = libDir.listFiles();
                if(libs != null) {
                    for (File lib : libs) {

                        urls.add(lib.toURI().toURL());

                    }
                }
            }
            return urls.toArray(new URL[urls.size()]);
        } catch (MalformedURLException e) {
            throw new RuntimeException(e);
        }
    }
}


